# coding=utf-8
import json
import datetime

from django.contrib.auth.models import User
from django.test import Client
from django.test import TestCase
from django.urls import reverse

from web.views.notifications import get_today_midnight_date
from web.api_views.subject import get_subjects_order, get_subjects_filtered, serialize_subject, SUBJECT_LIST_GENERIC, \
    SUBJECT_LIST_NO_VISIT, SUBJECT_LIST_REQUIRE_CONTACT
from web.models import Subject, Appointment, Visit
from web.tests.functions import create_subject, create_worker, create_get_suffix, create_visit, create_appointment


class TestApi(TestCase):
    def setUp(self):
        self.subject = create_subject()
        self.client = Client()
        username = 'piotr'
        password = 'top_secret'
        self.user = User.objects.create_user(
            username=username, email='jacob@bla', password=password)
        self.worker = create_worker(self.user)
        self.client.login(username=username, password=password)

    def test_cities(self):
        city_name = "some city"

        response = self.client.get(reverse('web.api.cities'))
        self.assertEqual(response.status_code, 200)

        cities = json.loads(response.content)['cities']

        self.assertFalse(city_name in cities)

        self.subject.city = city_name
        self.subject.save()

        response = self.client.get(reverse('web.api.cities'))
        cities = json.loads(response.content)['cities']

        self.assertTrue(city_name in cities)

    def test_referrals(self):
        referral_name = "some referral"

        response = self.client.get(reverse('web.api.referrals'))
        self.assertEqual(response.status_code, 200)

        referrals = json.loads(response.content)['referrals']

        self.assertFalse(referral_name in referrals)

        self.subject.referral = referral_name
        self.subject.save()

        response = self.client.get(reverse('web.api.referrals'))
        referrals = json.loads(response.content)['referrals']

        self.assertTrue(referral_name in referrals)

    def test_subjects_general(self):
        response = self.client.get(reverse('web.api.subjects', kwargs={'type': SUBJECT_LIST_GENERIC}))
        self.assertEqual(response.status_code, 200)

    def test_subjects_no_visit(self):
        response = self.client.get(reverse('web.api.subjects', kwargs={'type': SUBJECT_LIST_NO_VISIT}))
        self.assertEqual(response.status_code, 200)

    def test_subjects_require_contact(self):
        response = self.client.get(reverse('web.api.subjects', kwargs={'type': SUBJECT_LIST_REQUIRE_CONTACT}))
        self.assertEqual(response.status_code, 200)

    def test_subjects_invalid(self):
        response = self.client.get(reverse('web.api.subjects', kwargs={'type': "bla"}))
        self.assertEqual(response.status_code, 500)

    def test_subjects_general_search(self):
        name = "Piotrek"
        self.subject.first_name = name
        self.subject.save()

        params = {
            "columns[0][search][value]": "another_name",
            "columns[0][data]": "first_name"
        }
        url = ("%s" + create_get_suffix(params)) % reverse('web.api.subjects', kwargs={'type': SUBJECT_LIST_GENERIC})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertFalse(name in response.content)

        params["columns[0][search][value]"] = name
        url = ("%s" + create_get_suffix(params)) % reverse('web.api.subjects', kwargs={'type': SUBJECT_LIST_GENERIC})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(name in response.content)

    def check_subject_filtered(self, filters, result):
        subjects = get_subjects_filtered(Subject.objects.all(), filters)
        self.assertEqual(len(result), subjects.count())
        for index in range(len(result)):
            self.assertEqual(result[index], subjects[index])

    def check_subject_ordered(self, order, result):
        subjects = get_subjects_order(Subject.objects.all(), order, "asc")
        self.assertEqual(len(result), subjects.count())
        for index in range(len(result)):
            self.assertEqual(result[index], subjects[index])

        subjects = get_subjects_order(Subject.objects.all(), order, "desc")
        length = len(result)
        self.assertEqual(length, subjects.count())
        for index in range(length):
            self.assertEqual(result[length - index - 1], subjects[index])

    def test_subjects_sort_nd_number(self):
        subject = self.subject
        subject.nd_number = "PPP"
        subject.save()

        subject2 = create_subject(2)
        subject2.nd_number = "QQQ"
        subject2.save()

        self.check_subject_ordered("nd_number", [subject, subject2])

    def test_subjects_sort_id(self):
        subject = self.subject

        subject2 = create_subject(2)

        self.check_subject_ordered("id", [subject, subject2])

    def test_subjects_sort_date_born(self):
        subject = self.subject
        subject.date_born = get_today_midnight_date()
        subject.save()

        subject2 = create_subject(2)
        subject2.date_born = get_today_midnight_date() + datetime.timedelta(days=1)
        subject2.save()

        self.check_subject_ordered("date_born", [subject, subject2])

    def test_subjects_sort_default_location(self):
        subject = self.subject

        self.check_subject_ordered("default_location", [subject])

    def test_subjects_sort_screening_number(self):
        subject = self.subject
        subject.screening_number = "PPP"
        subject.save()

        subject2 = create_subject(2)
        subject2.screening_number = "QQQ"
        subject2.save()

        self.check_subject_ordered("screening_number", [subject, subject2])

    def test_subjects_sort_last_name(self):
        subject = self.subject
        subject.last_name = "XXX"
        subject.save()

        subject2 = create_subject(2)
        subject2.last_name = "YYY"
        subject2.save()

        self.check_subject_ordered("last_name", [subject, subject2])

    def test_subjects_sort_dead(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        subject2 = create_subject(2)
        subject2.dead = False
        subject2.save()

        self.check_subject_ordered("dead", [subject2, subject])

    def test_subjects_sort_resigned(self):
        subject = self.subject
        subject.resigned = True
        subject.save()

        subject2 = create_subject(2)
        subject2.resigned = False
        subject2.save()

        self.check_subject_ordered("resigned", [subject2, subject])

    def test_subjects_sort_postponed(self):
        subject = self.subject
        subject.postponed = True
        subject.save()

        subject2 = create_subject(2)
        subject2.postponed = False
        subject2.save()

        self.check_subject_ordered("postponed", [subject2, subject])

    def test_subjects_filter_dead(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        subject2 = create_subject(2)
        subject2.dead = False
        subject2.save()

        self.check_subject_filtered([["dead", "true"]], [subject])
        self.check_subject_filtered([["dead", "false"]], [subject2])

    def test_subjects_filter_nd_number(self):
        subject = self.subject
        subject.nd_number = "PPP"
        subject.save()

        subject2 = create_subject(2)
        subject2.nd_number = "QQQ"
        subject2.save()

        self.check_subject_filtered([["nd_number", "P"]], [subject])

    def test_subjects_filter_screening_number(self):
        subject = self.subject
        subject.screening_number = "PPP"
        subject.save()

        subject2 = create_subject(2)
        subject2.screening_number = "QQQ"
        subject2.save()

        self.check_subject_filtered([["screening_number", "Q"]], [subject2])

    def test_subjects_filter_last_name(self):
        subject = self.subject
        subject.last_name = "XXX"
        subject.save()

        subject2 = create_subject(2)
        subject2.last_name = "YYY"
        subject2.save()

        self.check_subject_filtered([["last_name", "Q"]], [])

    def test_subjects_filter_resigned(self):
        subject = self.subject
        subject.resigned = True
        subject.save()

        subject2 = create_subject(2)
        subject2.resigned = False
        subject2.save()

        self.check_subject_filtered([["resigned", "true"]], [subject])
        self.check_subject_filtered([["resigned", "false"]], [subject2])

    def test_subjects_filter_postponed(self):
        subject = self.subject
        subject.postponed = True
        subject.save()

        subject2 = create_subject(2)
        subject2.postponed = False
        subject2.save()

        self.check_subject_filtered([["postponed", "true"]], [subject])
        self.check_subject_filtered([["postponed", "false"]], [subject2])

    def test_subjects_filter_default_location(self):
        subject = self.subject

        self.check_subject_filtered([["default_location", str(subject.default_location.id)]], [subject])
        self.check_subject_filtered([["default_location", "-1"]], [])

    def test_subjects_filter_unknown(self):
        subject = self.subject

        self.check_subject_filtered([["some_unknown", "unknown data"]], [subject])
        self.check_subject_filtered([["", ""]], [subject])

    def test_serialize_subject(self):
        subject = self.subject
        subject.dead = True

        subject_json = serialize_subject(subject)
        self.assertEqual("YES", subject_json["dead"])

    def test_subjects_filter_visit_1_DONE(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        visit = create_visit(subject)
        appointment = create_appointment(visit)
        appointment.status = Appointment.APPOINTMENT_STATUS_FINISHED
        appointment.save()
        visit.mark_as_finished()

        self.check_subject_filtered([["visit_1", "DONE"]], [subject])
        self.check_subject_filtered([["visit_2", "DONE"]], [])

        self.check_subject_filtered([["visit_1", "MISSED"]], [])
        self.check_subject_filtered([["visit_1", "EXCEED"]], [])
        self.check_subject_filtered([["visit_1", "IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "SHOULD_BE_IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "UPCOMING"]], [])

    def test_subjects_filter_visit_1_MISSED(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        visit = create_visit(subject)
        appointment = create_appointment(visit)
        appointment.status = Appointment.APPOINTMENT_STATUS_CANCELLED
        appointment.save()
        visit.mark_as_finished()

        self.check_subject_filtered([["visit_1", "MISSED"]], [subject])
        self.check_subject_filtered([["visit_2", "MISSED"]], [])

        self.check_subject_filtered([["visit_1", "DONE"]], [])
        self.check_subject_filtered([["visit_1", "EXCEED"]], [])
        self.check_subject_filtered([["visit_1", "IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "SHOULD_BE_IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "UPCOMING"]], [])

    def test_subjects_filter_visit_1_EXCEED(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        visit = create_visit(subject)
        visit.datetime_begin = get_today_midnight_date() + datetime.timedelta(days=-365 * 2)
        visit.datetime_end = get_today_midnight_date() + datetime.timedelta(days=-365 * 1)
        visit.save()

        self.check_subject_filtered([["visit_1", "EXCEED"]], [subject])
        self.check_subject_filtered([["visit_2", "EXCEED"]], [])

        self.check_subject_filtered([["visit_1", "DONE"]], [])
        self.check_subject_filtered([["visit_1", "MISSED"]], [])
        self.check_subject_filtered([["visit_1", "IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "SHOULD_BE_IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "UPCOMING"]], [])

    def test_subjects_filter_visit_1_IN_PROGRESS(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        visit = create_visit(subject)
        appointment = create_appointment(visit)
        appointment.status = Appointment.APPOINTMENT_STATUS_SCHEDULED
        appointment.save()

        self.check_subject_filtered([["visit_1", "IN_PROGRESS"]], [subject])
        self.check_subject_filtered([["visit_2", "IN_PROGRESS"]], [])

        self.check_subject_filtered([["visit_1", "DONE"]], [])
        self.check_subject_filtered([["visit_1", "MISSED"]], [])
        self.check_subject_filtered([["visit_1", "EXCEED"]], [])
        self.check_subject_filtered([["visit_1", "SHOULD_BE_IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "UPCOMING"]], [])

    def test_subjects_filter_visit_1_SHOULD_BE_IN_PROGRESS(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        visit = create_visit(subject)

        self.check_subject_filtered([["visit_1", "SHOULD_BE_IN_PROGRESS"]], [subject])
        self.check_subject_filtered([["visit_2", "SHOULD_BE_IN_PROGRESS"]], [])

        self.check_subject_filtered([["visit_1", "DONE"]], [])
        self.check_subject_filtered([["visit_1", "MISSED"]], [])
        self.check_subject_filtered([["visit_1", "EXCEED"]], [])
        self.check_subject_filtered([["visit_1", "IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "UPCOMING"]], [])

    def test_subjects_filter_visit_1_UPCOMING(self):
        subject = self.subject
        subject.dead = True
        subject.save()

        visit = create_visit(subject)
        visit.datetime_begin = get_today_midnight_date() + datetime.timedelta(days=2)
        visit.datetime_end = get_today_midnight_date() + datetime.timedelta(days=20)
        visit.save()

        self.check_subject_filtered([["visit_1", "UPCOMING"]], [subject])
        self.check_subject_filtered([["visit_2", "UPCOMING"]], [])

        self.check_subject_filtered([["visit_1", "DONE"]], [])
        self.check_subject_filtered([["visit_1", "MISSED"]], [])
        self.check_subject_filtered([["visit_1", "EXCEED"]], [])
        self.check_subject_filtered([["visit_1", "IN_PROGRESS"]], [])
        self.check_subject_filtered([["visit_1", "SHOULD_BE_IN_PROGRESS"]], [])

    def test_subjects_filter_visit_1_visit_2_combined(self):
        subject = self.subject
        subject.save()

        visit = create_visit(subject)
        appointment = create_appointment(visit)
        appointment.status = Appointment.APPOINTMENT_STATUS_FINISHED
        appointment.save()
        visit.mark_as_finished()

        self.check_subject_filtered([["visit_1", "DONE"]], [subject])
        self.check_subject_filtered([["visit_2", "UPCOMING"]], [subject])

        self.check_subject_filtered([["visit_1", "DONE"], ["visit_2", "UPCOMING"]], [subject])
        self.check_subject_filtered([["visit_1", "UPCOMING"], ["visit_2", "DONE"]], [])
